# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "17/12/2018"


from xml.etree.ElementTree import TreeBuilder, Element, ElementTree
from collections import defaultdict
from itertools import count, chain
from typing import Union
import json
import pprint
import base64
import pickle
import logging
import typing
from .node import Node
from .link import Link
from ast import literal_eval

_logger = logging.getLogger(__name__)


class Scheme(object):
    """
    class to define a workflow scheme from nodes and links

    :param typing.Iterable nodes: set of Node contained in this scheme.
                                  note: subschemes are considered as Node.
                                  So if any defined it should be given as a
                                  Node
    :param typing.Iterable links: set of Link
    :type: typing.Iterable

    """

    _JSON_DESCRIPTION = "description"
    _JSON_TITLE = "title"
    _JSON_NODES = "nodes"
    _JSON_LINKS = "links"

    def __init__(
        self,
        nodes: typing.Iterable = None,
        links: typing.Iterable = None,
        description: str = None,
        title: str = None,
    ):
        self.__rnodes = None
        self.__rlinks = None
        self.__rsub_schemes = None

        self._reset(nodes=nodes, links=links, description=description, title=title)

    @property
    def rnodes(self) -> Union[None, list]:
        """All the nodes recursively. So this mean contained in this scheme or
        in a subscheme"""
        if self.__rnodes is None:
            self.__rnodes = list(self.nodes)
            for subscheme in self.sub_schemes:
                self.__rnodes.extend(subscheme.rnodes)
        return self.__rnodes

    @property
    def rlinks(self) -> Union[None, dict]:
        """all the links recursively. So this mean contained in this scheme or
        in a subscheme"""
        if self.__rlinks is None:
            self.__rlinks = self.links
            for subscheme in self.sub_schemes:
                self.__rlinks.update(subscheme.rlinks)
        return self.__rlinks

    @property
    def sub_schemes(self) -> Union[None, list]:
        """list of sub schemes contained by this scheme"""
        return self.__sub_schemes

    @property
    def rsub_schemes(self) -> Union[None, list]:
        """list of all sub schemes contained in this scheme recursively"""
        if self.__rsub_schemes is None:
            self.__rsub_schemes = self.sub_schemes
            for sub_scheme in self.sub_schemes:
                self.__rsub_schemes.extend(sub_scheme.rsub_schemes)
        return self.__rsub_schemes

    def _reset(self, nodes, links, description, title) -> None:
        self.title = title or ""
        self.description = description or ""
        # clear structure
        self.links = {}  # keys are link ID, values are Link
        self.__sub_schemes = []
        self.__rnodes = None
        self.__rlinks = None

        if links is not None:
            for link in links:
                self.links[link.id] = link
        self.nodes = nodes or []
        """list of nodes"""
        self.nodes_dict = {}
        """dict with node id as key and node as value"""

        # register all sub scheme
        for node in self.nodes:
            if isinstance(node, SubScheme):
                self.__sub_schemes.append(node)
                self.nodes_dict[node.id] = node
        # update nodes for sub_schemes
        for node in self.rnodes:
            self.nodes_dict[node.id] = node

        if links is not None:
            self._update_nodes_from_links()

    def final_nodes(self) -> list:
        """

        :return: list of final nodes (with no output) and which hasn't any
                 control node upstream
        """
        res = []
        for node in self.rnodes:
            assert isinstance(node, Node)
            if node.isfinal():
                res.append(node)
        return res

    def start_nodes(self) -> list:
        """

        :return: list of nodes starting the workflow. Those does not require
                 any input_data.
        :rtype: list
        """
        res = []
        for node in self.rnodes:
            assert isinstance(node, Node)
            if node.isstart():
                res.append(node)
        return res

    def endlessNodes(self) -> list:
        """

        :return: list of final nodes.
        :rtype: list
        """
        res = []
        for node in self.rnodes:
            assert isinstance(node, Node)
            if node.endless is True:
                res.append(node)
        return res

    def save_to(self, output_file: str) -> None:
        """
        Save the scheme as an xml formated file to `stream`

        :param output_file: name of the output file.
        :type: str
        """
        if output_file.lower().endswith(".json"):
            self.save_as_json(output_file)
        else:
            self.save_as_xml(output_file)

    def save_as_xml(self, output_file: str) -> None:
        """
        save current scheme to a default xml format

        :param str output_file: file path
        """
        tree = self.scheme_to_etree(data_format="literal")
        indent(tree.getroot(), 0)

        tree.write(output_file)

    def save_as_json(self, output_file: str) -> None:
        """
        save current scheme to a default json format

        :param str output_file: file path
        """
        with open(output_file, "w") as json_file:
            json.dump(self.to_json(), json_file)

    def nodes_to_json(self) -> list:
        """

        :return: nodes to json compatible format
        :rtype: list
        """
        res = []
        for node in self.nodes:
            res.append(node.to_json())
        return res

    @staticmethod
    def nodes_from_json(json_data: dict) -> tuple:
        """

        :param json_data: data containing the json definition
        :return: nodes, sub_schemes
        :rtype: tuple
        :raises: ValueError if the json is not a valid Node description
        """
        nodes = []
        sub_schemes = []
        for node_json_data in json_data:
            # if contains a node description then this is a subscheme
            if Scheme._JSON_NODES in node_json_data:
                sub_scheme = SubScheme.load_from_json(node_json_data)
                sub_schemes.append(sub_scheme)
            else:
                nodes.append(Node.from_json(node_json_data))
        return nodes, sub_schemes

    def links_to_json(self) -> list:
        """

        :return: links to json compatible format
        :rtype: list
        """
        res = []
        for link in self.links.values():
            res.append(link.to_json())
        return res

    @staticmethod
    def links_from_json(json_data: dict) -> list:
        """

        :param json_data: data containing the json definition
        :return: list of Link defined by the json data
        :rtype: list
        """
        links = []
        for link_json_data in json_data:
            links.append(Link.from_json(link_json_data))
        return links

    def to_json(self) -> dict:
        """
        Convert scheme to json

        :return: json dict
        :rtype: dict
        """
        return {
            self._JSON_DESCRIPTION: self.description,
            self._JSON_TITLE: self.title,
            self._JSON_NODES: self.nodes_to_json(),
            self._JSON_LINKS: self.links_to_json(),
        }

    @staticmethod
    def from_json_file(json_file_path: str):
        """
        Create and load Scheme from a json file

        :param str json_file_path: json file
        :return: Scheme fitting the json description
        :rtype: Scheme
        :raises: ValueError if file not found or invalid
        """
        scheme = Scheme()
        try:
            scheme.load_from_json_file(json_file_path)
        except ValueError as e:
            _logger.error(e)
            return None
        else:
            return scheme

    def load_from_json_file(self, json_file_path: str):
        """

        :param str json_file_path: path to the json file containing the scheme
                                   description

        :return: Scheme fitting the json description contains if the file.
                 If description is incomplete, return None
        :rtype: Union[Scheme, None]
        """
        try:
            with open(json_file_path, "r") as json_file:
                json_data = json.load(json_file)
        except IOError as e:
            _logger.error("fail to read json file", str(e))
        else:
            self.load_from_json(json_data=json_data)

    @staticmethod
    def load_scheme_info_from_json(json_data: dict) -> tuple:
        """
        load fom json stream the Scheme Information

        :param json_data: scheme description
        :return: nodes, links, sub-schemes, title, description
        :rtype: tuple
        """
        # load title
        if Scheme._JSON_TITLE not in json_data:
            _logger.warning("no title found in the json")
            title = None
        else:
            title = json_data[Scheme._JSON_TITLE]
        # load description
        if Scheme._JSON_DESCRIPTION not in json_data:
            _logger.warning("no description found in the json")
            description = None
        else:
            description = json_data[Scheme._JSON_DESCRIPTION]
        # load links
        if Scheme._JSON_LINKS not in json_data:
            _logger.error("no link found in the json")
            links = None
        else:
            try:
                links = Scheme.links_from_json(json_data=json_data[Scheme._JSON_LINKS])
            except ValueError as e:
                _logger.error(e)
                links = None
        # load nodes
        nodes, sub_schemes = None, None
        if Scheme._JSON_NODES not in json_data:
            _logger.error("no nodes found in the json")
        else:
            try:
                nodes, sub_schemes = Scheme.nodes_from_json(
                    json_data=json_data[Scheme._JSON_NODES]
                )
            except ValueError as e:
                _logger.error(e)
                nodes = None
        return nodes, links, sub_schemes, title, description

    def load_from_json(self, json_data: dict):
        """

        :param json_data: scheme description
        :raise ValueError: if sink or source channel missing or if link id
                           missing or if sink or source node missing
        """
        nodes, links, sub_schemes, title, description = self.load_scheme_info_from_json(
            json_data
        )
        # create scheme if possible
        if nodes is None or links is None:
            raise ValueError(
                "unable to load scheme from json description." "Information missing"
            )
        else:
            if sub_schemes is not None:
                nodes.extend(sub_schemes)
            self._reset(nodes=nodes, links=links, description=description, title=title)

    def scheme_to_etree(
        self, data_format: str = "literal", pickle_fallback: bool = False
    ):
        """
        Return an 'xml.etree.ElementTree' representation of the scheme.
        """
        builder = TreeBuilder(element_factory=Element)
        builder.start(
            "scheme",
            {
                "version": "2.0",
                "title": self.title or "",
                "description": self.description or "",
            },
        )

        # Nodes
        node_ids = defaultdict(count().__next__)
        builder.start("nodes", {})
        for node in self.nodes:  # type: SchemeNode
            attrs = {"id": node.id, "qualified_name": node._qualified_name}

            if type(node) is not Node:
                attrs["scheme_node_type"] = "%s.%s" % (
                    type(node).__name__,
                    type(node).__module__,
                )
            builder.start("node", attrs)
            builder.end("node")

        builder.end("nodes")

        # Links
        link_ids = defaultdict(count().__next__)
        builder.start("links", {})
        for link in self.links:  # type: SchemeLink
            source = link.source_node_id
            sink = link.sink_node_id
            source_id = node_ids[source]
            sink_id = node_ids[sink]
            attrs = {
                "id": str(link_ids[link]),
                "source_node_id": str(source_id),
                "sink_node_id": str(sink_id),
                "source_channel": link.source_channel,
                "sink_channel": link.sink_channel,
                "enabled": "true" if link.enabled else "false",
            }
            builder.start("link", attrs)
            builder.end("link")

        builder.end("links")

        # Annotations
        annotation_ids = defaultdict(count().__next__)
        builder.start("thumbnail", {})
        builder.end("thumbnail")

        # Node properties/settings
        builder.start("node_properties", {})
        for node in self.nodes:
            data = None
            if node.properties:
                try:
                    data, format = dumps(
                        node.properties,
                        format=data_format,
                        pickle_fallback=pickle_fallback,
                    )
                except Exception:
                    _logger.error(
                        "Error serializing properties for node %r",
                        node.title,
                        exc_info=True,
                    )
                if data is not None:
                    builder.start(
                        "properties", {"node_id": str(node_ids[node]), "format": format}
                    )
                    builder.data(data)
                    builder.end("properties")

        builder.end("node_properties")

        builder.end("scheme")
        root = builder.close()
        tree = ElementTree(root)
        return tree

    def _update_nodes_from_links(self):
        """
        Update upstream and downstream nodes from links definition
        """
        self._clear_nodes_connections()
        for link_id, link in self.links.items():
            source_node = self.nodes_dict[link.source_node_id]
            sink_node = self.nodes_dict[link.sink_node_id]
            source_node.downstream_nodes.add(self.nodes_dict[link.sink_node_id])
            sink_node.upstream_nodes.add(self.nodes_dict[link.source_node_id])

    def _clear_nodes_connections(self):
        """
        clear for all nodes downstream and upstream nodes
        """
        for node in self.nodes:
            assert isinstance(node, Node)
            node.downstream_nodes = set()
            node.upstream_nodes = set()

    def has_final_join(self):
        """
        :return: True if we need to send a 'end' signal before closing the
                 workflow. This is needed in the 'acquisition workflow' like
                 tomwer and the DataWatcher process for example.
        :rtype: bool
        """
        for node in self.nodes:
            if node.need_stop_join:
                return True
        return False

    @staticmethod
    def from_desc(desc):
        """

        :param desc:
        :return: instance of Scheme from it description.
        :rtype: :class:`Scheme`
        """
        nodes = []
        nodes_dict = {}

        for node_d in desc.nodes:
            node = Node(id=node_d.id, processing_pt=node_d.qualified_name)
            nodes.append(node)
            nodes_dict[node.id] = node
            data = node_d.data
            if data:
                properties = loads(data.data, data.format)
                node.properties = properties
            else:
                node.properties = {}
            node.qualified_name = node_d.qualified_name

        for link_d in desc.links:
            upstream_node = nodes_dict[link_d.source_node_id]
            assert isinstance(upstream_node, Node)
            downstream_node = nodes_dict[link_d.sink_node_id]
            upstream_node.downstream_nodes.add(downstream_node)
            downstream_node.upstream_nodes.add(upstream_node)

        scheme = Scheme(nodes=nodes, links=desc.links)
        scheme.title = desc.title
        scheme.description = desc.description

        return scheme

    def load_handlers(self):
        """
        load all nodes handlers.
        """
        for node in self.nodes:
            node.load_handlers()


class SubScheme(Scheme, Node):
    """
    Define a sub-scheme of the workflow (or subworkflow). SubScheme are
    as Scheme expect that they are not 'root'.

    :param nodes: set of Node
    :type: typing.Iterable
    :param links: set of Node
    :type: typing.Iterable
    :param description: description of the subscheme
    :type: str
    :param error_handler: ErrorHandler
    """

    def __init__(
        self,
        nodes: typing.Iterable,
        links: typing.Iterable,
        description: str = None,
        error_handler=None,
        title="",
        id=None,
    ):
        Node.__init__(self, processing_pt=None, error_handler=error_handler, id=id)
        Scheme.__init__(
            self, nodes=nodes, links=links, title=title, description=description
        )

    def to_json(self):
        """
        Convert sub scheme to json

        :return: json dict
        :rtype: dict
        """
        desc = Scheme.to_json(self)
        desc.update(Node.to_json(self))
        return desc

    @staticmethod
    def load_from_json(json_data: dict):
        """

        :param json_data: scheme description.

        :raise ValueError: if sink or source channel missing or if link id
                           missing or if sink or source node missing.
        """
        (
            nodes,
            links,
            sub_schemes,
            title,
            description,
        ) = Scheme.load_scheme_info_from_json(json_data)
        _id, _properties, _process_pt = Node.load_node_info_from_json(json_data)
        if _properties is not None:
            _logger.warning("SubScheme properties are not managed")
        if _process_pt is not None:
            _logger.warning("SubScheme pointer to ptocess is not managed")
        # create scheme if possible
        if nodes is None or links is None:
            raise ValueError(
                "unable to load scheme from json description." "Information missing"
            )
        sub_scheme = SubScheme(
            nodes=nodes, links=links, description=description, title=title, id=_id
        )
        return sub_scheme


def contains_control_nodes(nodes_list: typing.Iterable):
    """
    Return the list of the 'control' nodes.

    :param typing.Iterable nodes_list:
    :return:
    """
    for _node in nodes_list:
        if _node.endless or contains_control_nodes(_node.upstream_nodes):
            return True
    return False


def loads(string: str, format: str) -> object:
    """load stream from the given format.

    :param str string: stream to load
    :param str format: format
    :return: object contained in the stream
    :rtype: object
    """
    if format == "literal":
        return literal_eval(string)
    elif format == "json":
        return json.loads(string)
    elif format == "pickle":
        return pickle.loads(base64.decodebytes(string.encode("ascii")))
    else:
        raise ValueError("Unknown format")


def indent(element, level=0, indent="\t"):
    """
    Indent an instance of a :class:`Element`. Based on
    (http://effbot.org/zone/element-lib.htm#prettyprint).

    """

    def empty(text):
        return not text or not text.strip()

    def indent_(element, level, last):
        child_count = len(element)

        if child_count:
            if empty(element.text):
                element.text = "\n" + indent * (level + 1)

            if empty(element.tail):
                element.tail = "\n" + indent * (level + (-1 if last else 0))

            for i, child in enumerate(element):
                indent_(child, level + 1, i == child_count - 1)

        else:
            if empty(element.tail):
                element.tail = "\n" + indent * (level + (-1 if last else 0))

    return indent_(element, level, True)


def dumps(obj, format="literal", prettyprint=False, pickle_fallback=False):
    """
    Serialize `obj` using `format` ('json' or 'literal') and return its
    string representation and the used serialization format ('literal',
    'json' or 'pickle').

    If `pickle_fallback` is True and the serialization with `format`
    fails object's pickle representation will be returned

    """
    if format == "literal":
        try:
            return (literal_dumps(obj, prettyprint=prettyprint, indent=1), "literal")
        except (ValueError, TypeError) as ex:
            if not pickle_fallback:
                raise

            _logger.debug("Could not serialize to a literal string")

    elif format == "json":
        try:
            return (json.dumps(obj, indent=1 if prettyprint else None), "json")
        except (ValueError, TypeError):
            if not pickle_fallback:
                raise

            _logger.debug("Could not serialize to a json string")

    elif format == "pickle":
        return base64.encodebytes(pickle.dumps(obj)).decode("ascii"), "pickle"

    else:
        raise ValueError("Unsupported format %r" % format)

    if pickle_fallback:
        _logger.warning("Using pickle fallback")
        return base64.encodebytes(pickle.dumps(obj)).decode("ascii"), "pickle"
    else:
        raise Exception("Something strange happened.")


# This is a subset of PyON serialization.
def literal_dumps(obj, prettyprint=False, indent=4):
    """
    Write obj into a string as a python literal.
    """
    memo = {}
    NoneType = type(None)

    def check(obj):
        if type(obj) in [int, float, bool, NoneType, str, bytes]:
            return True

        if id(obj) in memo:
            raise ValueError("{0} is a recursive structure".format(obj))

        memo[id(obj)] = obj

        if type(obj) in [list, tuple]:
            return all(map(check, obj))
        elif type(obj) is dict:
            return all(map(check, chain(iter(obj.keys()), iter(obj.values()))))
        else:
            raise TypeError(
                "{0} can not be serialized as a python " "literal".format(type(obj))
            )

    check(obj)

    if prettyprint:
        return pprint.pformat(obj, indent=indent)
    else:
        return repr(obj)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2015-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""test representation is well taken into account"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "02/04/2020"


import unittest
from pypushflow.representation import Scheme, Node, Link
from pypushflow.Workflow import ProcessableWorkflow


def exec_(scheme, input_=None):
    """
    Simple execution procedure of a workflow.

    :param Scheme scheme:
    :param input_: workflow input if any
    """
    assert isinstance(scheme, ProcessableWorkflow)
    scheme._start_actor.trigger(input_)
    scheme._end_actor.join()
    return scheme._end_actor.out_data


class TestScheme(unittest.TestCase):
    """Test the Scheme class"""

    def setUp(self):
        self.node1 = Node(processing_pt="pypushflow.test.utils.test_function")
        self.node2 = Node(processing_pt="pypushflow.test.utils.ProcessingClass1")
        self.node3 = Node(processing_pt="pypushflow.test.utils.ProcessingClass2")
        self.link1 = Link(self.node1, self.node2, "data", "data")
        self.link2 = Link(self.node2, self.node3, "data", "data")

        self.scheme = Scheme(
            nodes=(self.node1, self.node2, self.node3), links=(self.link1, self.link2)
        )

    def testSchemeDefinition(self):
        """simple test of scheme definition"""
        self.assertEqual(len(self.scheme.nodes), 3)
        self.assertEqual(len(self.scheme.links), 2)
        self.assertTrue(self.scheme.start_nodes() == [self.node1])
        self.assertTrue(self.scheme.final_nodes() == [self.node3])

    def testProcessing(self):
        processable_workflow = ProcessableWorkflow(scheme=self.scheme)
        channel_out, out_ = exec_(
            scheme=processable_workflow, input_=(None, {"data": 0})
        )
        self.assertEqual(out_["data"], 7)


class TestNodeExecution(unittest.TestCase):
    """Insure Node.execute works"""

    def testCase1(self):
        "test that a specific function from her name can be executed"
        node = Node(processing_pt="pypushflow.test.utils.test_function")
        res = node.execute(
            node.process_pt, properties={}, input_name="data", input_data={"data": 0}
        )
        self.assertEqual(res, (None, {"data": 2}))

    def testCase2(self):
        "test that an callable class can be executed from her name"
        node = Node(processing_pt="pypushflow.test.utils.ProcessingClass1")
        res = node.execute(
            node.process_pt, properties={}, input_name="data", input_data={"data": 0}
        )
        self.assertEqual(res, (None, {"data": 4}))

    def testCase3(self):
        """Test that a class with handler can be executed"""
        node = Node(processing_pt="pypushflow.test.utils.ProcessingClass2")
        res = node.execute(
            node.process_pt, properties={}, input_name="data", input_data={"data": 0}
        )
        self.assertEqual(res, (None, {"data": 1}))

    def testCase4(self):
        script = "pypushflow.test.pythonActorTest.run"
        node = Node(processing_pt=script)
        res = node.execute(
            node.process_pt,
            properties={},
            input_name="data",
            input_data={"name": "pythonActorTest"},
        )
        self.assertEqual(res, (None, {"reply": "Hello pythonActorTest!"}))


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestScheme, TestNodeExecution):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"

import logging
import multiprocessing

# from pypushflow import UtilsMongoDb

logger = logging.getLogger("pypushflow")


class StopActor(object):
    def __init__(self, parent=None, errorHandler=None, name="Stop actor"):
        self.errorHandler = errorHandler
        self.name = name
        self.lock = multiprocessing.Lock()
        self.lock.acquire()
        self.out_data = None
        self.parent = parent

    def trigger(self, inData):
        logger.debug(
            "In trigger {0}, errorHandler = {1}".format(self.name, self.errorHandler)
        )
        # if self.parent is not None and hasattr(self.parent, 'mongoId'):
        #     UtilsMongoDb.closeMongo(self.parent.mongoId)
        if self.errorHandler is not None:
            self.errorHandler.errorHandler.stopActor.trigger(inData)
        else:
            self.out_data = inData
            self.lock.release()

    def join(self, timeout=7200):
        self.lock.acquire(timeout=timeout)



__authors__ = ["Bioinformatics Laboratory, University of Ljubljana", "H.Payno"]
__license__ = "[GNU GPL v3+]: https://www.gnu.org/licenses/gpl-3.0.en.html"
__date__ = "29/05/2017"


from xml.etree.ElementTree import parse
from .parser import (
    _scheme,
    _node,
    _link,
    _nxNodeProperty,
    _nxRelation,
    _nxLink,
    _nxNode,
)
import logging
from .scheme import Node, Scheme
from .parser import Parser

logger = logging.getLogger(__name__)


class MomlParser(Parser):
    """This is a dictionnary to convert an OrangeWidget into a core process."""

    @staticmethod
    def scheme_load(file_, load_handlers=True):
        """

        :param file_: file containing the scheme definition
        :param bool load_handlers: try to load the handlers of each node. Used
                                   to make sure the process won't fail
        :return: :class:`Scheme`
        """
        desc = MomlParser.parse_moml_stream(file_)
        scheme = Scheme.from_desc(desc)
        if load_handlers is True:
            scheme.load_handlers()
        return scheme

    @staticmethod
    def parse_moml_stream(stream):
        doc = parse(stream)
        scheme = MomlParser.parse_moml_etree(doc)
        return scheme

    @staticmethod
    def parse_moml_etree(tree):
        def convertToOrangeLinks(nx_links, nx_relations):
            def getLink(_id):
                if _id not in links:
                    links[_id] = {"id": _id}
                return links[_id]

            links = {}
            # convert nx (.omlm) liks and relations to orange links
            for nx_link in nx_links:
                link = getLink(nx_link.relation)
                # TODO: for now some port / link type are not managed
                if nx_link.port in ("In", "Out", "No mesh defined"):
                    logger.warning(nx_link.port + " not managed yet")
                    continue
                else:
                    node_id, input_output = nx_link.port.split(".", -1)
                    if input_output.lower() in ("output", "other", "true"):
                        link["source_node_id"] = node_id
                    else:
                        link["sink_node_id"] = node_id

            for relation in nx_relations:
                assert relation.id in links
                links[relation.id]["source_channel"] = relation.class_
                links[relation.id]["sink_channel"] = relation.class_
                links[relation.id]["properties"] = relation.properties

            orangeLinks = []
            for linkid, link in links.items():
                # TODO: this condition is due from the case that some link are
                #  not managed yet.
                if "source_node_id" not in link or "sink_node_id" not in link:
                    continue
                l = _link(
                    id=link["id"],
                    source_node_id=link["source_node_id"],
                    sink_node_id=link["sink_node_id"],
                    source_channel=link["source_channel"],
                    sink_channel=link["sink_channel"],
                    enabled=True,
                )
                orangeLinks.append(l)

            return orangeLinks

        nodes, nx_links, nx_relations = [], [], []

        # Collect all nodes
        for node in tree.findall(".//entity"):
            node_id = node.get("name")
            node_class = node.get("class")
            node_properties = []
            for property in node.findall("property"):
                _property = _nxNodeProperty(
                    name=property.get("name"),
                    class_=property.get("class"),
                    value=property.get("value"),
                )
                node_properties.append(_property)
            nodes.append(
                _nxNode(
                    id=node_id,
                    class_=node_class,
                    properties=node_properties,
                    data=None,
                    qualified_name=node_class,
                )
            )

        # collect all nx links
        for link in tree.findall("link"):
            _my_link = _nxLink(port=link.get("port"), relation=link.get("relation"))
            nx_links.append(_my_link)

        # collect all nx relations
        for relation in tree.findall("relation"):
            relation_properties = []
            for property in relation.findall("property"):
                _property = _nxNodeProperty(
                    name=property.get("name"),
                    class_=property.get("class"),
                    value=property.get("value"),
                )
                relation_properties.append(_property)

            _relation = _nxRelation(
                id=relation.get("name"),
                class_=relation.get("class"),
                properties=relation_properties,
            )
            nx_relations.append(_relation)

        links = convertToOrangeLinks(nx_links=nx_links, nx_relations=nx_relations)

        return _scheme(
            version="nx 0.1",
            title=("undefined"),
            description=None,
            nodes=nodes,
            links=links,
            annotations=None,
        )



__authors__ = ["Bioinformatics Laboratory, University of Ljubljana", "H.Payno"]
__license__ = "[GNU GPL v3+]: https://www.gnu.org/licenses/gpl-3.0.en.html"
__date__ = "29/05/2017"


from xml.etree.ElementTree import parse
from .parser import (
    _scheme,
    _node,
    _link,
    _data,
    _annotation,
    _text_params,
    _arrow_params,
)
from ast import literal_eval
import ast
import logging
from .scheme import Scheme
from .parser import Parser

logger = logging.getLogger(__name__)


class OwsParser(Parser):
    """
    Parser managing the .ows (orange) files
    """

    @staticmethod
    def scheme_load(file_, load_handlers=True):
        """

        :param file_: file containing the scheme definition
        :param bool load_handlers: try to load the handlers of each node. Used
                                   to make sure the process won't fail
        :return: :class:`Scheme`
        """
        desc = OwsParser.parse_ows_stream(file_)
        scheme = Scheme.from_desc(desc)
        if load_handlers is True:
            scheme.load_handlers()
        return scheme

    @staticmethod
    def parse_ows_etree_v_2_0(tree):
        aliases = Parser.get_aliases()
        scheme = tree.getroot()
        nodes, links, annotations = [], [], []

        # First collect all properties
        properties = {}
        for property in tree.findall("node_properties/properties"):
            node_id = property.get("node_id")
            format = property.get("format")
            if "data" in property.attrib:
                data = property.get("data")
            else:
                data = property.text
            properties[node_id] = _data(format, data)

        # Collect all nodes
        for node in tree.findall("nodes/node"):
            node_id = node.get("id")
            qualified_name = node.get("qualified_name")
            if qualified_name in aliases:
                logger.info(
                    "replace"
                    + str(qualified_name)
                    + "by"
                    + str(aliases[qualified_name])
                )
                qualified_name = aliases[qualified_name]

            node = _node(
                id=node_id,
                title=node.get("title"),
                name=node.get("name"),
                position=tuple_eval(node.get("position", None)),
                project_name=node.get("project_name", None),
                qualified_name=qualified_name,
                version=node.get("version", ""),
                data=properties.get(node_id, None),
            )
            nodes.append(node)

        for link in tree.findall("links/link"):
            params = _link(
                id=link.get("id"),
                source_node_id=link.get("source_node_id"),
                sink_node_id=link.get("sink_node_id"),
                source_channel=link.get("source_channel"),
                sink_channel=link.get("sink_channel"),
                enabled=link.get("enabled") == "true",
            )
            links.append(params)

        for annot in tree.findall("annotations/*"):
            if annot.tag == "text":
                rect = tuple_eval(annot.get("rect", "(0.0, 0.0, 20.0, 20.0)"))

                font_family = annot.get("font-family", "").strip()
                font_size = annot.get("font-size", "").strip()

                font = {}
                if font_family:
                    font["family"] = font_family
                if font_size:
                    font["size"] = int(font_size)

                annotation = _annotation(
                    id=annot.get("id"),
                    type="text",
                    params=_text_params(rect, annot.text or "", font),
                )
            elif annot.tag == "arrow":
                start = tuple_eval(annot.get("start", "(0, 0)"))
                end = tuple_eval(annot.get("end", "(0, 0)"))
                color = annot.get("fill", "red")
                annotation = _annotation(
                    id=annot.get("id"),
                    type="arrow",
                    params=_arrow_params((start, end), color),
                )
            annotations.append(annotation)
        return _scheme(
            version=scheme.get("version"),
            title=scheme.get("title", ""),
            description=scheme.get("description"),
            nodes=nodes,
            links=links,
            annotations=annotations,
        )

    @staticmethod
    def parse_ows_stream(stream):
        doc = parse(stream)
        scheme_el = doc.getroot()
        version = scheme_el.get("version", None)
        if version is None:
            # Fallback: check for "widgets" tag.
            if scheme_el.find("widgets") is not None:
                version = "1.0"
            else:
                logger.warning("<scheme> tag does not have a 'version' attribute")
                version = "2.0"

        if version == "1.0":
            raise ValueError("old .ows version are not managed")
        elif version == "2.0":
            return OwsParser.parse_ows_etree_v_2_0(doc)
        else:
            raise ValueError("unrecognize scheme definition version")


# ---- TAKE back from Orange3 ---------


def tuple_eval(source):
    """
    Evaluate a python tuple literal `source` where the elements are
    constrained to be int, float or string. Raise ValueError if not
    a tuple literal.

    >>> tuple_eval("(1, 2, "3")")
    (1, 2, '3')

    """
    if source is None:
        return None
    node = ast.parse(source, "<source>", mode="eval")

    if not isinstance(node.body, ast.Tuple):
        raise ValueError("%r is not a tuple literal" % source)

    if not all(
        isinstance(el, (ast.Str, ast.Num)) or
        # allow signed number literals in Python3 (i.e. -1|+1|-1.0)
        (
            isinstance(el, ast.UnaryOp)
            and isinstance(el.op, (ast.UAdd, ast.USub))
            and isinstance(el.operand, ast.Num)
        )
        for el in node.body.elts
    ):
        raise ValueError("Can only contain numbers or strings")

    return literal_eval(source)


def resolve_replaced(scheme_desc, registry):
    widgets = registry.widgets()
    nodes_by_id = {}  # type: Dict[str, _node]
    replacements = {}
    replacements_channels = {}  # type: Dict[str, Tuple[dict, dict]]
    # collect all the replacement mappings
    for desc in widgets:  # type: WidgetDescription
        if desc.replaces:
            for repl_qname in desc.replaces:
                replacements[repl_qname] = desc.qualified_name

        input_repl = {}
        for idesc in desc.inputs or []:  # type: InputSignal
            for repl_qname in idesc.replaces or []:  # type: str
                input_repl[repl_qname] = idesc.name
        output_repl = {}
        for odesc in desc.outputs:  # type: OutputSignal
            for repl_qname in odesc.replaces or []:  # type: str
                output_repl[repl_qname] = odesc.name
        replacements_channels[desc.qualified_name] = (input_repl, output_repl)

    # replace the nodes
    nodes = scheme_desc.nodes
    for i, node in list(enumerate(nodes)):
        if (
            not registry.has_widget(node.qualified_name)
            and node.qualified_name in replacements
        ):
            qname = replacements[node.qualified_name]
            desc = registry.widget(qname)
            nodes[i] = node._replace(
                qualified_name=desc.qualified_name, project_name=desc.project_name
            )
        nodes_by_id[node.id] = nodes[i]

    # replace links
    links = scheme_desc.links
    for i, link in list(enumerate(links)):  # type: _link
        nsource = nodes_by_id[link.source_node_id]
        nsink = nodes_by_id[link.sink_node_id]

        _, source_rep = replacements_channels.get(nsource.qualified_name, ({}, {}))
        sink_rep, _ = replacements_channels.get(nsink.qualified_name, ({}, {}))

        if link.source_channel in source_rep:
            link = link._replace(source_channel=source_rep[link.source_channel])
        if link.sink_channel in sink_rep:
            link = link._replace(sink_channel=sink_rep[link.sink_channel])
        links[i] = link

    return scheme_desc._replace(nodes=nodes, links=links)



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"

from pypushflow.Workflow import Workflow
from pypushflow.Submodel import Submodel
from pypushflow.StopActor import StopActor
from pypushflow.StartActor import StartActor
from pypushflow.PythonActor import PythonActor
import unittest


class Workflow1(Workflow):
    """
    Workflow containing one start actor,
    one python actor and one stop actor.
    """

    def __init__(self, name):
        Workflow.__init__(self, name)
        self.startActor = StartActor(parent=self)
        self.pythonActor = PythonActor(
            parent=self,
            script="pypushflow.test.pythonActorTest.py",
            name="Python Actor Test",
        )
        self.stopActor = StopActor(parent=self)
        self.startActor.connect(self.pythonActor)
        self.pythonActor.connect(self.stopActor)

    def run(self, inData):
        self.startActor.trigger(inData)
        self.stopActor.join(timeout=15)
        return self.stopActor.outData


class Workflow2(Workflow):
    """
    Workflow with error handling, containing one start actor,
    one python actor and one stop actor.

    The python actor throws an exception.
    """

    def __init__(self, name):
        Workflow.__init__(self, name)
        self.startActor = StartActor(parent=self)
        self.pythonActor = PythonActor(
            parent=self,
            script="pypushflow.test.pythonErrorHandlerTest.py",
            name="Python Error Handler Test",
            errorHandler=self,
        )
        self.stopActor = StopActor(parent=self)
        self.startActor.connect(self.pythonActor)
        self.pythonActor.connect(self.stopActor)
        self.connectOnError(self.stopActor)

    def run(self, inData):
        self.startActor.trigger(inData)
        self.stopActor.join(timeout=5)
        return self.stopActor.outData


class Submodel1(Submodel):
    """
    Submodel containing one python actor.
    """

    def __init__(self, parent, name):
        Submodel.__init__(self, parent, name=name)
        self.pythonActor = PythonActor(
            parent=self,
            script="pypushflow.test.pythonActorTest.py",
            name="Python Actor Test",
        )
        self.getPort("In").connect(self.pythonActor)
        self.pythonActor.connect(self.getPort("Out"))


class Workflow3(Workflow):
    """
    Workflow containing one start actor,
    one submodel and one stop actor.
    """

    def __init__(self, name):
        Workflow.__init__(self, name)
        self.startActor = StartActor(parent=self)
        self.submodel1 = Submodel1(parent=self, name="Submodel 1")
        self.stopActor = StopActor(parent=self)
        self.startActor.connect(self.submodel1.getPort("In"))
        self.submodel1.getPort("Out").connect(self.stopActor)

    def run(self, inData):
        self.startActor.trigger(inData)
        self.stopActor.join(timeout=15)
        return self.stopActor.outData


class Submodel2(Submodel):
    """
    Submodel containing one python actor which throws an exception.
    """

    def __init__(self, parent, name):
        Submodel.__init__(self, parent, name=name)
        self.pythonActor = PythonActor(
            parent=self,
            script="pypushflow.test.pythonErrorHandlerTest.py",
            name="Python Error Handler Test",
            errorHandler=self,
        )
        self.getPort("In").connect(self.pythonActor)
        self.pythonActor.connect(self.getPort("Out"))


class Workflow4(Workflow):
    """
    Workflow containing one start actor,
    one submodel which throws an exception and one stop actor.
    """

    def __init__(self, name):
        Workflow.__init__(self, name)
        self.startActor = StartActor()
        self.submodel2 = Submodel2("Submodel 2")
        self.stopActor = StopActor()
        self.startActor.connect(self.submodel2.getPort("In"))
        self.submodel2.getPort("Out").connect(self.stopActor)

    def run(self, inData):
        self.startActor.trigger(inData)
        self.stopActor.join(timeout=5)
        return self.stopActor.outData


def suite():
    test_suite = unittest.TestSuite()
    for ui in ():
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



#!/usr/bin/env python
# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2015-2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Unique place where the version number is defined.

provides:
* version = "1.2.3" or "1.2.3-beta4"
* version_info = named tuple (1,2,3,"beta",4)
* hexversion: 0x010203B4
* strictversion = "1.2.3b4
* debianversion = "1.2.3~beta4"
* calc_hexversion: the function to transform a version_tuple into an integer

This is called hexversion since it only really looks meaningful when viewed as the
result of passing it to the built-in hex() function.
The version_info value may be used for a more human-friendly encoding of the same information.

The hexversion is a 32-bit number with the following layout:
Bits (big endian order)     Meaning
1-8     PY_MAJOR_VERSION (the 2 in 2.1.0a3)
9-16     PY_MINOR_VERSION (the 1 in 2.1.0a3)
17-24     PY_MICRO_VERSION (the 0 in 2.1.0a3)
25-28     PY_RELEASE_LEVEL (0xA for alpha, 0xB for beta, 0xC for release candidate and 0xF for final)
29-32     PY_RELEASE_SERIAL (the 3 in 2.1.0a3, zero for final releases)

Thus 2.1.0a3 is hexversion 0x020100a3.

"""

from __future__ import absolute_import, print_function, division

__authors__ = ["Jérôme Kieffer", "H.Payno"]
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "17/06/2016"
__status__ = "production"
__docformat__ = "restructuredtext"
__all__ = [
    "date",
    "version_info",
    "strictversion",
    "hexversion",
    "debianversion",
    "calc_hexversion",
]

RELEASE_LEVEL_VALUE = {
    "dev": 0,
    "alpha": 10,
    "beta": 11,
    "gamma": 11,
    "rc": 12,
    "final": 15,
}

MAJOR = 0
MINOR = 2
MICRO = 0
RELEV = "rc"  # <16
SERIAL = 1  # <16

date = __date__

from collections import namedtuple

_version_info = namedtuple(
    "version_info", ["major", "minor", "micro", "releaselevel", "serial"]
)

version_info = _version_info(MAJOR, MINOR, MICRO, RELEV, SERIAL)

strictversion = version = debianversion = "%d.%d.%d" % version_info[:3]
if version_info.releaselevel != "final":
    version += "-%s%s" % version_info[-2:]
    debianversion += (
        "~adev%i" % version_info[-1] if RELEV == "dev" else "~%s%i" % version_info[-2:]
    )
    prerel = "a" if RELEASE_LEVEL_VALUE.get(version_info[3], 0) < 10 else "b"
    if prerel not in "ab":
        prerel = "a"
    strictversion += prerel + str(version_info[-1])


def calc_hexversion(major=0, minor=0, micro=0, releaselevel="dev", serial=0):
    """Calculate the hexadecimal version number from the tuple version_info:

    :param major: integer
    :param minor: integer
    :param micro: integer
    :param relev: integer or string
    :param serial: integer
    :return: integerm always increasing with revision numbers
    """
    try:
        releaselevel = int(releaselevel)
    except ValueError:
        releaselevel = RELEASE_LEVEL_VALUE.get(releaselevel, 0)

    hex_version = int(serial)
    hex_version |= releaselevel * 1 << 4
    hex_version |= int(micro) * 1 << 8
    hex_version |= int(minor) * 1 << 16
    hex_version |= int(major) * 1 << 24
    return hex_version


hexversion = calc_hexversion(*version_info)

if __name__ == "__main__":
    print(version)



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "16/10/2019"

import os
import bson
import datetime
import pymongo

from bson.objectid import ObjectId


def getDateTimeString():
    return datetime.datetime.now()


def getMongoUrl():
    return os.environ.get("PYPUSHFLOW_MONGOURL", None)


def initMongo(name):
    workflowId = None
    mongoUrl = getMongoUrl()
    if mongoUrl is not None:
        initiator = os.environ.get("PYPUSHFLOW_INITIATOR", "Unknown")
        host = os.environ.get("PYPUSHFLOW_HOST", "Unknown")
        port = os.environ.get("PYPUSHFLOW_PORT", "Unknown")
        objectId = os.environ.get("PYPUSHFLOW_OBJECTID", str(bson.objectid.ObjectId()))
        collection = pymongo.MongoClient(mongoUrl).pybes.pybes
        workflowData = {
            "_id": bson.objectid.ObjectId(objectId),
            "Request ID": objectId,
            "startTime": getDateTimeString(),
            "initiator": initiator,
            "host": host,
            "port": port,
            "name": name,
            "status": "started",
            "actors": [],
        }
        insertOneResult = collection.insert_one(workflowData)
        workflowId = insertOneResult.inserted_id
    return workflowId


def setMongoStatus(workflowId, status):
    mongoUrl = getMongoUrl()
    if mongoUrl is not None:
        collection = pymongo.MongoClient(mongoUrl).pybes.pybes
        dictWorkflow = collection.find_one({"_id": workflowId})
        dictWorkflow["status"] = status
        collection.update_one({"_id": workflowId}, {"$set": dictWorkflow}, upsert=False)


def closeMongo(workflowId):
    mongoUrl = getMongoUrl()
    if mongoUrl is not None:
        collection = pymongo.MongoClient(mongoUrl).pybes.pybes
        dictWorkflow = collection.find_one({"_id": workflowId})
        if dictWorkflow["status"] != "error":
            dictWorkflow["status"] = "finished"
        dictWorkflow["stopTime"] = getDateTimeString()
        collection.update_one({"_id": workflowId}, {"$set": dictWorkflow}, upsert=False)


def initActor(name, workflowId, actorData={}, script=None):
    actorId = None
    mongoUrl = getMongoUrl()
    if mongoUrl is not None:
        collection = pymongo.MongoClient(mongoUrl).pybes.pybes
        actorId = ObjectId()
        actor = {
            "_id": actorId,
            "startTime": getDateTimeString(),
            "name": name,
            "status": "started",
            "script": script,
        }
        actor.update(actorData)
        collection.update({"_id": workflowId}, {"$push": {"actors": actor}})
    return actorId


def addDataToActor(workflowId, actorId, actorData={}):
    mongoUrl = getMongoUrl()
    if mongoUrl is not None:
        collection = pymongo.MongoClient(mongoUrl).pybes.pybes
        dictWorkflow = collection.find_one({"_id": workflowId})
        for actor in dictWorkflow["actors"]:
            if actor["_id"] == actorId:
                actor.update(actorData)
                break
        collection.update_one({"_id": workflowId}, {"$set": dictWorkflow}, upsert=False)



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"


import logging

logger = logging.getLogger("pypushflow")


class Port(object):
    def __init__(self, errorHandler, name):
        self.name = errorHandler.name + "." + name
        self.errorHandler = errorHandler
        self.listActor = []
        self.inPortTrigger = None

    def connect(self, actor):
        logger.debug("Connect {0} -> actorName {1}".format(self.name, actor.name))
        self.listActor.append(actor)

    def setTrigger(self, trigger):
        self.inPortTrigger = trigger

    def trigger(self, *args, **kwargs):
        logger.debug("In {0} trigger".format(self.name))
        if len(self.listActor) > 0:
            for actor in self.listActor:
                logger.debug(
                    "In trigger {0} -> actorName {1}".format(
                        self.errorHandler.name, actor.name
                    )
                )
                actor.trigger(*args, **kwargs)
        if self.inPortTrigger is not None:
            logger.debug(
                "In {0} trigger, trigger = {1}".format(
                    self.errorHandler.name, self.inPortTrigger
                )
            )
            self.inPortTrigger(*args, **kwargs)


class Submodel(object):
    def __init__(self, parent, errorHandler=None, name=None, portNames=["In", "Out"]):
        self.parent = parent
        # self.mongoId = self.parent.mongoId
        self.name = name
        self.errorHandler = errorHandler
        self.dictPort = {}
        self.listOnErrorActor = []
        for portName in portNames:
            self.dictPort[portName] = Port(self, portName)

    def getActorPath(self):
        return self.parent.getActorPath() + "/" + self.name.replace("%", " ")

    def getPort(self, portName):
        logger.debug("In {0} getPort, portName = {1}".format(self.name, portName))
        return self.dictPort[portName]

    def connect(self, actor, portName="Out"):
        logger.debug(
            "In {0} connect, portName = {2} -> actorName = {1}".format(
                self.name, actor.name, portName
            )
        )
        self.dictPort[portName].connect(actor)

    def connectOnError(self, actor):
        logger.debug(
            "In connectOnError in subModule {0}, actor name {1}".format(
                self.name, actor.name
            )
        )
        self.listOnErrorActor.append(actor)

    def triggerOnError(self, *args, **kwargs):
        for onErrorActor in self.listOnErrorActor:
            logger.debug(
                "In triggerOnError in subModule {0}, trigger actor {1}, inData = {2}".format(
                    self.name, onErrorActor.name, args[0]
                )
            )
            onErrorActor.trigger(*args, **kwargs)
        if self.errorHandler is not None:
            self.errorHandler.triggerOnError(*args, **kwargs)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "19/08/2019"

from collections import namedtuple
import importlib
import logging

_logger = logging.getLogger(__name__)


class Parser(object):
    @staticmethod
    def scheme_load(file_, load_handlers=True):
        """

        :param file_: file containing the scheme definition
        :param bool load_handlers: try to load the handlers of each node. Used
                                   to make sure the process won't fail
        :return: :class:`Scheme`
        """
        raise NotImplementedError()

    @staticmethod
    def get_aliases():
        """

        :return: the list of aliases for a processing_pt (module.class or
                 module.function name)
        :rtype: dict
        """
        aliases = {}
        try:
            import ppfaddon
        except ImportError:
            return aliases
        else:
            import pkgutil

            for importer, modname, ispkg in pkgutil.iter_modules(ppfaddon.__path__):
                try:
                    mod_name = ".".join((ppfaddon.__name__, modname, "aliases"))
                    module = importlib.import_module(mod_name)
                except ImportError:
                    _logger.warning(
                        modname + " does not fit the add-on design, skip it"
                    )
                else:
                    if hasattr(module, "aliases"):
                        new_aliases = getattr(module, "aliases")
                        if not isinstance(new_aliases, dict):
                            raise TypeError("aliases should be an instance of dict")
                        else:
                            aliases.update(new_aliases)
        return aliases

    __main__ = scheme_load


_scheme = namedtuple(
    "_scheme", ["title", "version", "description", "nodes", "links", "annotations"]
)

_node = namedtuple(
    "_node",
    [
        "id",
        "title",
        "name",
        "position",
        "project_name",
        "qualified_name",
        "version",
        "data",
    ],
)

_data = namedtuple("_data", ["format", "data"])

_link = namedtuple(
    "_link",
    [
        "id",
        "source_node_id",
        "sink_node_id",
        "source_channel",
        "sink_channel",
        "enabled",
    ],
)

_annotation = namedtuple("_annotation", ["id", "type", "params"])

_text_params = namedtuple("_text_params", ["geometry", "text", "font"])

_arrow_params = namedtuple("_arrow_params", ["geometry", "color"])

_nxNode = namedtuple("_node", ["id", "class_", "properties", "data", "qualified_name"])

_nxNodeProperty = namedtuple("_property", ["name", "class_", "value"])

_nxLink = namedtuple("_nxLink", ["port", "relation"])

_nxRelation = namedtuple("_relation", ["id", "class_", "properties"])



#
# Copyright (c) European Synchrotron Radiation Facility (ESRF)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__authors__ = ["O. Svensson"]
__license__ = "MIT"
__date__ = "28/05/2019"

import os
import pprint
import logging
import traceback
import functools
import multiprocessing
from multiprocessing.pool import Pool as _Pool
from pypushflow.AbstractActor import AbstractActor
from pypushflow.representation.scheme.node import Node


logger = logging.getLogger(__name__)


class WorkflowException(Exception):
    def __init__(self, errorMessage="", traceBack="", data={}, msg=None):
        super(WorkflowException, self).__init__(msg)
        self.errorMessage = errorMessage
        if data is None:
            data = {}
        self.data = data
        self.traceBack = traceBack


def trace_unhandled_exceptions(func):
    @functools.wraps(func)
    def wrapped_func(*args, **kwargs):
        try:
            out_data = func(*args, **kwargs)
        except Exception as e:
            errorMessage = "{0}".format(e)
            logger.exception(errorMessage)
            traceBack = traceback.format_exc()
            return WorkflowException(
                errorMessage=errorMessage, traceBack=traceBack, data=args[1]
            )
        return out_data

    return wrapped_func


#############################################################################
# Create no daemon processes
# See : https://stackoverflow.com/a/8963618
#


class NoDaemonProcess(multiprocessing.Process):
    # make 'daemon' attribute always return False
    def _get_daemon(self):
        return False

    def _set_daemon(self, value):
        pass

    daemon = property(_get_daemon, _set_daemon)


# We sub-class multiprocessing.pool.Pool instead of multiprocessing.Pool
# because the latter is only a wrapper function, not a proper class.


class Pool(_Pool):
    Process = NoDaemonProcess


#
#
#############################################################################


@trace_unhandled_exceptions
def _exec_node(name: str, channel_name: str, data: dict, properties: dict):
    """
    Execute a node from the name of the process, input of the process and
    properties of the process

    :param str name: full name of the process to execute
    :param data: data to process
    :param properties: process properties / settings
    :return: result of Node.execute
    """
    logger.debug(
        "processing {0} on channel {1} with input {2} and {3} as "
        "properties".format(str(name), str(channel_name), str(data), str(properties))
    )
    return Node.execute(
        process_pt=name, properties=properties, input_data=data, input_name=channel_name
    )


class AsyncFactory:
    """
    TODO
    """

    def __init__(self, node, callback=None, errorCallback=None):
        self.node = node
        self.callback = callback
        self.errorCallback = errorCallback
        self.pool = Pool(1)

    def call(self, *args, **kwargs):
        logger.debug(
            "Before apply_async, func={0}, callback={1}, errorCallback={2}".format(
                self.node, self.callback, self.errorCallback
            )
        )
        logger.debug("args={0}, kwargs={1}".format(args, kwargs))
        if self.callback is None:
            self.pool.apply_async(_exec_node, args, kwargs)
        elif self.errorCallback is None:
            self.pool.apply_async(_exec_node, args, kwargs, self.callback)
        else:
            self.pool.apply_async(
                _exec_node, args, kwargs, self.callback, self.errorCallback
            )
        logger.debug("After apply_async")

    def wait(self):
        self.pool.close()
        self.pool.join()


class ActorWrapper(object):
    """
    TODO
    """

    def __init__(self, node):
        self.node = node

    @trace_unhandled_exceptions
    def run(self, in_data):
        logger.debug("In actor wrapper for {node}".format(node=self.node))
        output_channel_name, out_data = self.node.execute(in_data)
        if isinstance(out_data, WorkflowException):
            return output_channel_name, out_data
        else:
            in_data.update(out_data)
            return output_channel_name, out_data


class PythonActor(AbstractActor):
    """
    TODO

    * find a way to avoid 'duplication' of input 'script/process_pt'
      (should be done upstream)
    * I don't think script should have a default value.

    :param parent:
    :param name:
    :param errorHandler:
    :param script: script originally used
    :param node: Node from representation

    """

    def __init__(
        self,
        parent=None,
        name="Python Actor",
        errorHandler=None,
        script=None,
        node=None,
    ):
        if node is not None:
            if script is not None:
                raise ValueError(
                    "if a process pointer is provided, you "
                    "shouldn't provide a script"
                )
        elif script is None:
            raise ValueError("no script provide to the python actor")

        AbstractActor.__init__(self, parent=parent, name=name)
        self.error_handler = errorHandler
        self.list_error_handler = []
        # Import script
        self.script = script
        if script is not None:
            # module = importlib.import_module(os.path.splitext(script)[0])
            if script.endswith(".py"):
                script = "".join(os.path.splitext(script)[:-1])
            node = Node(".".join((script, "run")))
            self.actor_wrapper = ActorWrapper(node=node)
        else:
            self.actor_wrapper = ActorWrapper(node=node)

        self.in_data = None
        self.out_data = None
        self.async_factory = None

    def get_input_channel_name(self, type_):
        return self.actor_wrapper.node.get_input_channel_name(type_)

    def get_output_channel_name(self, type_):
        return self.actor_wrapper.node.get_output_channel_name(type_)

    def connectOnError(self, errorHandler):
        self.list_error_handler.append(errorHandler)

    def trigger(self, in_data):
        """
        'callback' function when this function is triggered.

        :param data: input data
        """
        channel, in_data = in_data
        logger.info("On trigger channel is " + str(channel))
        # cast data to dict if necessary
        if hasattr(in_data, "to_dict"):
            in_data = in_data.to_dict()

        self.in_data = in_data
        logger.debug(
            "In trigger {0}, inData = {1}".format(self.name, pprint.pformat(in_data))
        )
        if isinstance(in_data, WorkflowException):
            logger.error(
                "Error from previous actor! Not running actor {0}".format(self.name)
            )
            if self.error_handler is not None:
                workflowException = in_data
                oldInData = workflowException.data
                exceptionDict = {
                    "errorMessage": workflowException.errorMessage,
                    "traceBack": workflowException.traceBack.split("\n"),
                }
                oldInData["WorkflowException"] = exceptionDict
                self.triggerOnError(oldInData)

        self.async_factory = AsyncFactory(
            self.actor_wrapper.run,
            callback=self.triggerDownStreamActor,
            errorCallback=self.error_handler,
        )
        self.async_factory.call(
            self.actor_wrapper.node.process_pt,
            channel,
            in_data,
            self.actor_wrapper.node.properties,
        )

    def errorHandler(self, exception):
        logger.error("Error when running actor {0}!".format(self.name))
        workflowException = WorkflowException(
            errorMessage=exception, traceBack=None, data=None
        )
        inData = {"WorkflowException": workflowException}
        logger.error(exception)
        for errorHandler in self.list_error_handler:
            errorHandler.trigger(inData)
        if self.error_handler is not None:
            self.error_handler.triggerOnError(inData)

    def triggerDownStreamActor(self, output_last_processing=(None, {})):
        logger.warning("---------------------")
        logger.warning(output_last_processing)
        logger.warning("---------------------")
        try:
            output_channel, inData = output_last_processing
        except TypeError:
            output_channel, inData = None, output_last_processing
        logger.info(
            "In triggerDownStreamActor for {0}, Output channel is {1}, "
            "inData is {2}".format(self.name, output_channel, inData)
        )
        if isinstance(inData, WorkflowException):
            logger.error(
                "Error from previous actor! Not running down stream actors {0}".format(
                    [actor.name for actor in self.listDownStreamActor]
                )
            )
            workflowException = inData
            oldInData = workflowException.data
            exceptionDict = {
                "errorMessage": workflowException.errorMessage,
                "traceBack": workflowException.traceBack.split("\n"),
            }
            logger.warning(
                "oldInData type: {}, value: {}".format(type(oldInData), oldInData)
            )
            oldInData["WorkflowException"] = exceptionDict
            for errorHandler in self.list_error_handler:
                errorHandler.trigger((None, oldInData))
            if self.error_handler is not None:
                logger.error(
                    'Trigger on error on errorHandler "{0}"'.format(
                        self.error_handler.name
                    )
                )
                self.error_handler.triggerOnError(inData=(None, oldInData))
        else:
            out_data = {}
            if inData is not None:
                for key, value in inData.items():
                    if key in self.in_data:
                        if self.in_data[key] != value:
                            out_data[key] = value
                    else:
                        out_data[key] = value

            for downStreamActor in self.listDownStreamActor:
                downStreamActor.trigger((output_channel, inData))
